# -*- coding: utf-8 -*-
#
#    Monk is an unobtrusive data modeling, manipulation and validation library.
#    Copyright © 2011—2013  Andrey Mikhaylenko
#
#    This file is part of Monk.
#
#    Monk is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Monk is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with Monk.  If not, see <http://gnu.org/licenses/>.
"""
Compatibility
=============

This module is intended to hide away implementation details of various Python
versions.
"""
import sys
import types


func_types = (
    types.FunctionType,
    types.MethodType,
    # CPython: datetime.datetime.[utc]now()
    types.BuiltinFunctionType,
)


if sys.version_info < (3,0):
    text_type = unicode
    binary_type = str
else:
    text_type = str
    binary_type = bytes


def safe_str(value):
    """ Returns:

    * a `str` instance (bytes) in Python 2.x, or
    * a `str` instance (Unicode) in Python 3.x.

    """
    if sys.version_info < (3,0) and isinstance(value, unicode):
        return value.encode('utf-8')
    else:
        return str(value)


def safe_unicode(value):
    """ Returns:

    * a `unicode` instance in Python 2.x, or
    * a `str` instance in Python 3.x.

    """
    if sys.version_info < (3,0):
        if isinstance(value, str):
            return value.decode('utf-8')
        else:
            return unicode(value)
    else:
        return str(value)



# -*- coding: utf-8 -*-
#
#    Monk is an unobtrusive data modeling, manipulation and validation library.
#    Copyright © 2011—2013  Andrey Mikhaylenko
#
#    This file is part of Monk.
#
#    Monk is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Monk is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with Monk.  If not, see <http://gnu.org/licenses/>.
"""
MongoDB integration
~~~~~~~~~~~~~~~~~~~

This module combines Monk's modeling and validation capabilities with MongoDB.

Declaring indexes
-----------------

Let's declare a model with indexes::

    from monk.mongo import Document

    class Item(Document):
        structure = dict(text=unicode, slug=unicode)
        indexes = dict(text=None, slug=dict(unique=True))

Now create a model instance::

    item = Item(text=u'foo', slug=u'bar')

Save it and make sure the indexes are created::

    item.save(db)

The last line is roughly equivalent to::

    collection = db[item.collection]
    collection.ensure_index('text')
    collection.ensure_index('slug', unique=True)
    collection.save(dict(item))  # also validation, transformation, etc.

"""
from functools import partial

from bson import DBRef
from monk import modeling


class MongoResultSet(object):
    """ A wrapper for pymongo cursor that wraps each item using given function
    or class.

    .. warning::

       This class does not introduce caching.
       Iterating over results exhausts the cursor.

    """
    def __init__(self, cursor, wrapper):
        self._cursor = cursor
        self._wrap = wrapper

    def __iter__(self):
        return (self._wrap(x) for x in self._cursor)

    def __getitem__(self, index):
        return self._wrap(self._cursor[index])

    def __getattr__(self, attr):
        return getattr(self._cursor, attr)

    def ids(self):
        """ Returns a generator with identifiers of objects in set.
        These expressions are equivalent::

            ids = (item.get_id() for item in result_set)

            ids = result_set.ids()

        .. warning::

           This method **exhausts** the cursor, so an attempt to iterate over
           results after calling this method will *fail*. The results are *not*
           cached.

        """
        return (item.get_id() for item in self)

#    def count(self):
#        return self._cursor.count()


class MongoBoundDictMixin(object):
    """ Adds MongoDB-specific features to the dictionary.

    .. attribute:: collection

        Collection name.

    .. attribute:: indexes

        (TODO)

    """
    collection = None
    indexes = {}

    def __hash__(self):
        """ Collection name and id together make the hash; document class
        doesn't matter.

        Raises `TypeError` if collection or id is not set.
        """
        if self.collection and self.get_id():
            return hash(self.collection) | hash(self.get_id())
        raise TypeError('Document is unhashable: collection or id is not set')

    def __eq__(self, other):
        # both must inherit to this class
        if not isinstance(other, MongoBoundDictMixin):
            return False
        # both must have collections defined
        if not self.collection or not other.collection:
            return False
        # both must have ids
        if not self.get_id() or not other.get_id():
            return False

        # collections must be equal
        if self.collection != other.collection:
            return False
        # ids must be equal
        if self.get_id() != other.get_id():
            return False

        return True

    def __ne__(self, other):
        # this is required to override the call to dict.__eq__()
        return not self.__eq__(other)

    @classmethod
    def _ensure_indexes(cls, db):
        for field, kwargs in cls.indexes.items():
            kwargs = kwargs or {}
            db[cls.collection].ensure_index(field, **kwargs)

    @classmethod
    def wrap_incoming(cls, data, db):
        # XXX self.structure belongs to StructuredDictMixin !!
        return cls(dict_from_db(cls.structure, data, db))

    @classmethod
    def find(cls, db, *args, **kwargs):
        """
        Returns a :class:`MongoResultSet` object.

        Example::

            items = Item.find(db, {'title': u'Hello'})

        .. note::

           The arguments are those of pymongo collection's `find` method.
           A frequent error is to pass query key/value pairs as keyword
           arguments. This is **wrong**. In most cases you will want to pass
           a dictionary ("query spec") as the first positional argument.

        """
        cls._ensure_indexes(db)
        docs = db[cls.collection].find(*args, **kwargs)
        return MongoResultSet(docs, partial(cls.wrap_incoming, db=db))

    @classmethod
    def get_one(cls, db, *args, **kwargs):
        """
        Returns an object that corresponds to given query or ``None``.

        Example::

            item = Item.get_one(db, {'title': u'Hello'})

        """
        data = db[cls.collection].find_one(*args, **kwargs)
        if data:
            return cls.wrap_incoming(data, db)
        else:
            return None

    def save(self, db):
        """
        Saves the object to given database. Usage::

            item = Item(title=u'Hello')
            item.save(db)

        Collection name is taken from :attr:`MongoBoundDictMixin.collection`.
        """
        assert self.collection

        self._ensure_indexes(db)

        # XXX self.structure belongs to StructuredDictMixin !!
        outgoing = dict(dict_to_db(self, self.structure))

        object_id = db[self.collection].save(outgoing)

        if self.get('_id') is None:
            self['_id'] = object_id
        else:
            pass

        return object_id

    def get_id(self):
        """ Returns object id or ``None``.
        """
        return self.get('_id')

    def get_ref(self):
        """ Returns a `DBRef` for this object or ``None``.
        """
        _id = self.get_id()
        if _id is None:
            return None
        else:
            return DBRef(self.collection, _id)

    def remove(self, db):
        """
        Removes the object from given database. Usage::

            item = Item.get_one(db)
            item.remove(db)

        Collection name is taken from :attr:`MongoBoundDictMixin.collection`.
        """
        assert self.collection
        assert self.get_id()

        db[self.collection].remove(self.get_id())


def _db_to_dict_pairs(spec, data, db):
    for key, value in data.items():
        if isinstance(value, dict):
            yield key, dict(_db_to_dict_pairs(spec.get(key, {}), value, db))
        elif isinstance(value, DBRef):
            obj = db.dereference(value)
            cls = spec.get(key, dict)
            yield key, cls(obj, _id=obj['_id']) if obj else None
        else:
            yield key, value


def dict_from_db(spec, data, db):
    return dict(_db_to_dict_pairs(spec, data, db))


def _dict_to_db_pairs(spec, data):
    for key, value in data.items():
        if key == '_id' and value is None:
            # let the database assign an identifier
            continue
        if isinstance(value, dict):
            if '_id' in value:
                collection = spec[key].collection
                yield key, DBRef(collection, value['_id'])
            else:
                yield key, dict(_dict_to_db_pairs(spec.get(key, {}), value))
        else:
            yield key, value


def dict_to_db(data, spec={}):
    return dict(_dict_to_db_pairs(spec, data))


class Document(
        modeling.TypedDictReprMixin,
        modeling.DotExpandedDictMixin,
        modeling.StructuredDictMixin,
        MongoBoundDictMixin,
        dict
    ):
    """ A structured dictionary that is bound to MongoDB and supports dot
    notation for access to items.

    Inherits features from:

    * `dict` (builtin),
    * :class:`~TypedDictReprMixin`,
    * :class:`~DotExpandedDictMixin`,
    * :class:`~StructuredDictMixin` and
    * :class:`~MongoBoundDictMixin`.

    """
    def __init__(self, *args, **kwargs):
        super(Document, self).__init__(*args, **kwargs)
        self._insert_defaults()
        self._make_dot_expanded()

    def save(self, db):
        self.validate()
        return super(Document, self).save(db)



# -*- coding: utf-8 -*-
#
#    Monk is an unobtrusive data modeling, manipulation and validation library.
#    Copyright © 2011—2013  Andrey Mikhaylenko
#
#    This file is part of Monk.
#
#    Monk is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Monk is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with Monk.  If not, see <http://gnu.org/licenses/>.
"""
Data manipulation
~~~~~~~~~~~~~~~~~
"""
from monk.compat import text_type
from monk.schema import canonize


__all__ = [
    # functions
    'merge_defaults', 'merged',
    # helpers
    'unfold_list_of_dicts', 'unfold_to_list',
    # constants
    'TYPE_MERGERS',
    # merger functions
    'merge_any', 'merge_dict', 'merge_list',
]


# NOTE: updated at the end of file
TYPE_MERGERS = {}
""" The default set of type-specific value mergers:

* ``dict`` -- :func:`merge_dict`
* ``list`` -- :func:`merge_list`

"""


def merge_any(spec, value, mergers, fallback):
    """ Always returns the value as is.
    """
    return value


def merge_dict(spec, value, mergers, fallback):
    """ Returns a dictionary based on `value` with each value recursively
    merged with `spec`.
    """
    assert spec.datatype is dict

    if spec.optional and value is None:
        return None

    if spec.inner_spec is None:
        if value is None:
            return {}
        else:
            return value

    if value is not None and not isinstance(value, dict):
        # bogus value; will not pass validation but should be preserved
        return value

    data = value or {}
    result = {}

    for key in set(list(spec.inner_spec.keys()) + list(data.keys())):
        if key in spec.inner_spec:
            value = merge_defaults(spec.inner_spec[key], data.get(key),
                                   mergers, fallback)
        else:
            # never mind if there are nested structures: anyway we cannot check
            # them as they aren't in the spec
            value = data[key]

        result[key] = value

    return result


def merge_list(spec, value, mergers, fallback):
    """ Returns a list based on `value`:

    * missing required value is converted to an empty list;
    * missing required items are never created;
    * nested items are merged recursively.

    """
    assert spec.datatype is list

    item_spec = canonize(spec.inner_spec or None)

    if spec.optional and value is None:
        return None

    if not value:
        return []

    if value is not None and not isinstance(value, list):
        # bogus value; will not pass validation but should be preserved
        return value

    if item_spec.datatype is None:
        # any value is accepted as list item
        return value

    if item_spec.inner_spec:
        return [merge_defaults(item_spec.inner_spec, item, mergers, fallback)
                for item in value]

    return value


def merge_defaults(spec, value, mergers=TYPE_MERGERS, fallback=merge_any):
    """ Returns a copy of `value` recursively updated to match the `spec`:

    * New values are added whenever possible (including nested ones).
    * Existing values are never changed or removed.

      * Exception: container values (lists, dictionaries) may be populated;
        see respective merger functions for details.

    The result may not pass validation against the `spec`
    in the following cases:

    a) a required value is missing and the spec does not provide defaults;
    b) an existing value is invalid.

    The business logic is as follows:

    * if `value` is empty, use default value from `spec`;
    * if `value` is present or `spec` has no default value:

      * if `spec` datatype is present as a key in `mergers`,
        use the respective merger function to obtain the value;
      * if no merger is assigned to the datatype, use `fallback` function.

    See documentation on concrete merger functions for further details.

    :spec:
        A "natural" or "verbose" spec.

    :value:
        The value to merge into the `spec`.

    :mergers:
        An dictionary of merger functions assigned to specific types
        of values (sort of `{int: integer_merger_func}`).

        A merger function should accept the same arguments as this function,
        only with `spec` always being a :class:`~monk.schema.Rule` instance.

        Default: :attr:`TYPE_MERGERS`.

    :fallback:
        A merger function to use when no datatype-specific merger is found.

        Default: :func:`merge_any`

    Examples (with standard mergers)::

        >>> merge_defaults('foo', None)
        'foo'
        >>> merge_defaults('foo', 'bar')
        'bar'
        >>> merge_defaults({'a': 'foo'}, {})
        {'a': 'foo'}
        >>> merge_defaults({'a': [{'b': 123}]},
        ...                {'a': [{'b': None},
        ...                       {'x': 0}]})
        {'a': [{'b': 123}, {'b': 123, 'x': 0}]}

    """
    rule = canonize(spec)

    if value is None and rule.default is not None:
        return rule.default

    merger = mergers.get(rule.datatype, fallback)

    return merger(rule, value, mergers=mergers, fallback=fallback)


def merged(spec, data, mergers=TYPE_MERGERS):
    """
    .. deprecated:: 0.10.0

       Use :func:`merge_defaults` instead.
    """
    import warnings
    warnings.warn('merged() is deprecated, use merge_defaults() instead',
                  DeprecationWarning)

    return merge_defaults(spec, data, mergers=mergers)


class UNDEFINED:
    pass


def normalize_list_of_dicts(value, default_key, default_value=UNDEFINED):
    """
    Converts given value to a list of dictionaries as follows:

    * ``[{...}]`` → ``[{...}]``
    * ``{...}``   → ``[{...}]``
    * ``'xyz'``   → ``[{default_key: 'xyz'}]``
    * ``None``    → ``[{default_key: default_value}]``  (if specified)
    * ``None``    → ``[]``

    :param default_value:
        only Unicode, i.e. `str` in Python 3.x and **only** `unicode` in Python 2.x

    """
    if value is None:
        if default_value is UNDEFINED:
            return []
        value = default_value

    if isinstance(value, dict):
        return [value]

    if isinstance(value, text_type):
        return [{default_key: value}]

    if isinstance(value, list):
        if not all(isinstance(x, dict) for x in value):
            def _fix(x):
                return {default_key: x} if isinstance(x, text_type) else x
            return list(map(_fix, value))

    return value


def normalize_to_list(value):
    """
    Converts given value to a list  as follows:

    * ``[x]`` → ``[x]``
    * ``x``  → ``[x]``

    """
    if value and not isinstance(value, list):
        return [value]
    else:
        return value


def unfold_list_of_dicts(value, default_key):
    """
    .. deprecated:: 0.10.0

       Use :func:`normalize_list_of_dicts` instead.
    """
    return normalize_list_of_dicts(value, default_key)


def unfold_to_list(value):
    """
    .. deprecated:: 0.10.0

       Use :func:`normalize_to_list` instead.
    """
    return normalize_to_list(value)

TYPE_MERGERS.update({
    dict: merge_dict,
    list: merge_list,
})



# -*- coding: utf-8 -*-
#
#    Monk is an unobtrusive data modeling, manipulation and validation library.
#    Copyright © 2011—2013  Andrey Mikhaylenko
#
#    This file is part of Monk.
#
#    Monk is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Monk is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with Monk.  If not, see <http://gnu.org/licenses/>.
"""
Modeling
~~~~~~~~

DB-agnostic helpers to build powerful ODMs.

"""
from __future__ import unicode_literals

from .compat import text_type
from . import manipulation, validation


__all__ = ['DotExpandedDictMixin', 'DotExpandedDict', 'make_dot_expanded',
           'TypedDictReprMixin',
           'StructuredDictMixin']


def make_dot_expanded(data):
    if isinstance(data, DotExpandedDictMixin):
        return data
    elif isinstance(data, dict):
        pairs = []
        for key, value in data.items():
            pairs.append((key, make_dot_expanded(value)))
        return DotExpandedDict(pairs)
    elif isinstance(data, list):
        return [make_dot_expanded(x) for x in data]
    return data


class DotExpandedDictMixin(object):
    """ Makes the dictionary dot-expandable by exposing dictionary members
    via ``__getattr__`` and ``__setattr__`` in addition to ``__getitem__`` and
    ``__setitem__``. For example, this is the default API::

        data = {'foo': {'bar': 0 } }
        print data['foo']['bar']
        data['foo']['bar'] = 123

    This mixin adds the following API::

        print data.foo.bar
        data.foo.bar = 123

    Nested dictionaries are converted to dot-expanded ones on adding.
    """
    def _make_dot_expanded(self):
        for key, value in self.items():
            self[key] = make_dot_expanded(value)

    def __getattr__(self, attr):
        if not attr.startswith('_') and attr in self:
            return self[attr]
        raise AttributeError('Attribute or key {0.__class__.__name__}.{1} '
                             'does not exist'.format(self, attr))

    def __setattr__(self, attr, value):
        if not attr.startswith('_') and attr in self:
            self[attr] = value
        else:
            # Ambigous intent: cannot tell whether user wants to create
            # a dictionary key or actually set an object attribute.
            # Assuming the last option.
            super(DotExpandedDictMixin, self).__setattr__(attr, value)

    def __setitem__(self, key, value):
        if isinstance(value, dict) and \
           not isinstance(value, DotExpandedDict):
            value = make_dot_expanded(value)
        super(DotExpandedDictMixin, self).__setitem__(key, value)


class DotExpandedDict(DotExpandedDictMixin, dict):
    def __init__(self, *args, **kwargs):
        super(DotExpandedDict, self).__init__(*args, **kwargs)
        self._make_dot_expanded()


class TypedDictReprMixin(object):
    """ Makes ``repr(self)`` depend on ``unicode(self)``.
    """
    def __repr__(self):
        return '<{0.__class__.__name__} {1}>'.format(self, text_type(self))

    def __unicode__(self):
        return text_type(dict(self))

    def __str__(self):
        return text_type(dict(self))


class StructuredDictMixin(object):
    """ A dictionary with structure specification and validation.

    .. attribute:: structure

        The document structure specification. For details see
        :func:`monk.validation.validate`.

    """
    structure = {}
    #defaults = {}
    #required = []
    #validators = {}
    #with_skeleton = True

    def _insert_defaults(self):
        """ Inserts default values from :attr:`StructuredDictMixin.structure`
        to `self` by merging the two structures
        (see :func:`monk.manipulation.merged`).
        """
        with_defaults = manipulation.merged(self.structure, self)
        self.update(with_defaults)

    def validate(self):
        validation.validate(self.structure, self)



# coding: utf-8
#
#    Monk is an unobtrusive data modeling, manipulation and validation library.
#    Copyright © 2011—2013  Andrey Mikhaylenko
#
#    This file is part of Monk.
#
#    Monk is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Monk is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with Monk.  If not, see <http://gnu.org/licenses/>.
"""
Exceptions
~~~~~~~~~~
"""
class ValidationError(Exception):
    "Raised when a document or its part cannot pass validation."


class StructureSpecificationError(ValidationError):
    "Raised when malformed document structure is detected."


class MissingValue(ValidationError):
    """ Raised when the value is `None` and the rule neither allows this
    (i.e. a `datatype` is defined) nor provides a `default` value.
    """


class MissingKey(ValidationError):
    """ Raised when a dictionary key is defined in :attr:`Rule.inner_spec`
    but is missing from the value.
    """


class UnknownKey(ValidationError):
    """ Raised whan the value dictionary contains a key which is not
    in the dictionary's :attr:`Rule.inner_spec`.
    """



# -*- coding: utf-8 -*-
#
#    Monk is an unobtrusive data modeling, manipulation and validation library.
#    Copyright © 2011—2013  Andrey Mikhaylenko
#
#    This file is part of Monk.
#
#    Monk is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Monk is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with Monk.  If not, see <http://gnu.org/licenses/>.
"""
Validators
~~~~~~~~~~
"""
from monk.errors import ValidationError


def validate_choice(choices):
    """
    Checks whether the value belongs to given set of choices
    ::

        >>> check = validate_choice(['a', 'c'])
        >>> check('a')
        >>> check('b')
        Traceback (most recent call last):
        ...
        monk.errors.ValidationError: expected one of ['a', 'c'], got 'b'

    """
    def _validate_choice(value):
        if value not in choices:
            raise ValidationError('expected one of {0}, got {1!r}'.format(choices, value))
    return _validate_choice


def validate_range(start, stop):
    """
    Checks whether the numeric value belongs to given range
    ::

        >>> check = validate_range(0,1)
        >>> check(0)
        >>> check(1)
        >>> check(2)
        Traceback (most recent call last):
        ...
        monk.errors.ValidationError: expected value in range 0..1, got 2

    """
    def _validate_range(value):
        if not start <= value <= stop:
            raise ValidationError('expected value in range {0}..{1}, got {2!r}'.format(start, stop, value))
    return _validate_range


def validate_length(expected):
    """
    Checks whether the value is of given length
    ::

        >>> check = validate_length(2)
        >>> check('ab')
        >>> check([1, 2])
        >>> check('abc')
        Traceback (most recent call last):
        ...
        monk.errors.ValidationError: expected value of length 2, got 'abc'

    """
    def _validate_length(value):
        if len(value) != expected:
            raise ValidationError('expected value of length {0}, got {1!r}'.format(expected, value))
    return _validate_length




# -*- coding: utf-8 -*-
#
#    Monk is an unobtrusive data modeling, manipulation and validation library.
#    Copyright © 2011—2013  Andrey Mikhaylenko
#
#    This file is part of Monk.
#
#    Monk is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Monk is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with Monk.  If not, see <http://gnu.org/licenses/>.
"""
Schema Definition
~~~~~~~~~~~~~~~~~
"""
from . import compat, errors, validators


__all__ = [
    'Rule', 'canonize',
    # shortcuts:
    'any_value', 'any_or_none', 'optional', 'in_range', 'one_of'
]


#-------------------------------------------
# Classes
#

class Rule:
    """
    Extended specification of a field.  Allows marking it as optional.

    :param datatype:
        one of:

        * a type/class (stands for "an instance of this type/class")
        * `None` (stands for "any value of any kind or no value at all")

    :param default:
        an instance of `datatype` to be used in the absence of a value.

    :param inner_spec:
        a spec for nested data (can be another :class:`Rule` instance
        or something that :func:`canonize` can convert to one).

    :param optional:
        ``bool``; if ``True``, :class:`~monk.validation.MissingValue`
        is never raised.  Default is ``False``.

    :param dict_allow_unknown_keys:
        ``bool``; if ``True``, :class:`~monk.validation.UnknownKey`
        is never raised.  Default is ``False``.

    :param validators:
        a list of callables.

    .. note:: Missing Value vs. Missing Key vs. Unknown Key

       :MissingValue:
            Applies: to any value on any level.

            Trigger: the value is `None` and the rule neither allows this
            (i.e. a `datatype` is defined) nor provides a `default` value.

            Suppress: turn the `optional` setting on.
            This allows the value to be `None` even if a `datatype` is defined.

       :MissingKey:
            Applies: to a dictionary key.

            Trigger: a key is missing from the dictionary.

            Suppress: turn the `optional` setting on.
            This allows the key to be completely missing from an outer dictionary.

       :UnknownKey:
            Applies: to a dictionary key.

            Trigger: the dictionary contains a key which is not in
            the dictionary's `inner_spec`.

            Suppress: turn the `dict_allow_unknown_keys` setting on
            (of course on dictionary level).

    """
    def __init__(self, datatype, inner_spec=None, optional=False,
                 dict_allow_unknown_keys=False, default=None, validators=None):
        if isinstance(datatype, type(self)):
            raise ValueError('Cannot use a Rule instance as datatype')
        self.datatype = datatype
        self.inner_spec = inner_spec
        self.optional = optional
        self.dict_allow_unknown_keys = dict_allow_unknown_keys
        self.default = default
        self.validators = validators or []

        # sanity checks

        if default is not None and self.datatype and not isinstance(default, self.datatype):
            raise TypeError('Default value must match datatype {0} (got {1})'.format(
                self.datatype.__name__, default))

#        if self.inner_spec and not isinstance(self.inner_spec, self.datatype):
#            raise TypeError('Inner spec must match datatype {0} (got {1})'.format(
#                self.datatype.__name__, inner_spec))

    def __repr__(self):
        return '<Rule {datatype}{optional}{default}{inner_spec}{allow_unknown_keys}>'.format(
            datatype=('any' if self.datatype is None else
                self.datatype.__name__),
            optional=(' optional' if self.optional else ' required'),
            allow_unknown_keys=(' dict:allow-unknown-keys' if self.dict_allow_unknown_keys else ''),
            default=(' default={0}'.format(self.default)
                     if self.default is not None else ''),
            inner_spec=(' inner_spec={0}'.format(self.inner_spec)
                     if self.inner_spec is not None else ''))

    def __eq__(self, other):
        if (isinstance(other, type(self)) and self.__dict__ == other.__dict__):
            return True


#-------------------------------------------
# Functions
#

def canonize(spec, rule_kwargs={}):
    """
    Returns the canonic representation of given natural spec.

    :param spec: :term:`natural spec` (a `dict`)

    :return: :term:`detailed spec` (`dict` with :class:`Rule` instances as values)
    """
    value = spec

    if isinstance(value, Rule):
        rule = value
    elif value is None:
        rule = Rule(None, **rule_kwargs)
    elif isinstance(value, type):
        rule = Rule(value, **rule_kwargs)
    elif type(value) in compat.func_types:
        real_value = value()
        kwargs = dict(rule_kwargs, default=real_value)
        rule = Rule(datatype=type(real_value), **kwargs)
    elif isinstance(value, list):
        if value == []:
            # no inner spec, just an empty list as the default value
            kwargs = dict(rule_kwargs, default=value)
            rule = Rule(datatype=list, **kwargs)
        elif len(value) == 1:
            # the only item as spec for each item of the collection
            kwargs = dict(rule_kwargs, inner_spec=value[0])
            rule = Rule(datatype=list, **kwargs)
        else:
            raise errors.StructureSpecificationError(
                'Expected a list containing exactly 1 item; '
                'got {cnt}: {spec}'.format(cnt=len(value), spec=value))
    elif isinstance(value, dict):
        kwargs = dict(rule_kwargs, inner_spec=value)
        rule = Rule(datatype=dict, **kwargs)
    else:
        kwargs = dict(rule_kwargs, default=value)
        rule = Rule(datatype=type(value), **kwargs)

    return rule


#-------------------------------------------
# Shortcuts
#

def optional(spec):
    """
    Returns a canonized `spec` marked as optional.
    ::

        >>> optional(str) == Rule(datatype=str, optional=True)
        True

    """
    if isinstance(spec, Rule):
        spec.optional = True
        return spec
    else:
        return canonize(spec, rule_kwargs={'optional': True})


any_value = Rule(None)
"A shortcut for ``Rule(None)``"


any_or_none = Rule(None, optional=True)
"A shortcut for ``Rule(None, optional=True)``"


def one_of(choices, first_is_default=False):
    """
    A shortcut for a rule with :func:`~monk.validators.validate_choice` validator.
    ::

        choices = ['foo', 'bar']

        # these expressions are equal:

        one_of(choices)

        Rule(str, validators=[monk.validators.validate_choice(choices)])

        # default value can be taken from the first choice:

        one_of(choices, first_is_default=True)

        Rule(str, default=choices[0],
             validators=[monk.validators.choice(choices)])

    """
    assert choices

    if first_is_default:
        default_choice = choices[0]
    else:
        default_choice = None

    return Rule(datatype=type(choices[0]),
                default=default_choice,
                validators=[validators.validate_choice(choices)])


def in_range(start, stop, first_is_default=False):
    """
    A shortcut for a rule with :func:`~monk.validators.validate_range` validator.
    ::

        # these expressions are equal:

        in_range(0, 200)

        Rule(str, validators=[monk.validators.validate_range(0, 200)])

        # default value can be taken from the first choice:

        in_range(0, 200, first_is_default=True)

        Rule(str, default=0,
             validators=[monk.validators.validate_range(0, 200)])

    """
    if first_is_default:
        default_value = start
    else:
        default_value = None

    return Rule(datatype=int,
                default=default_value,
                validators=[validators.validate_range(start, stop)])



# -*- coding: utf-8 -*-
#
#    Monk is an unobtrusive data modeling, manipulation and validation library.
#    Copyright © 2011—2013  Andrey Mikhaylenko
#
#    This file is part of Monk.
#
#    Monk is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Monk is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with Monk.  If not, see <http://gnu.org/licenses/>.
"""
Helpers
=======
"""


def walk_dict(data):
    """ Generates pairs ``(keys, value)`` for each item in given dictionary,
    including nested dictionaries. Each pair contains:

    `keys`
        a tuple of 1..n keys, e.g. ``('foo',)`` for a key on root level or
        ``('foo', 'bar')`` for a key in a nested dictionary.
    `value`
        the value of given key or ``None`` if it is a nested dictionary and
        therefore can be further unwrapped.
    """
    assert hasattr(data, '__getitem__')
    for key, value in data.items():
        if isinstance(value, dict):
            yield (key,), None
            for keys, value in walk_dict(value):
                path = (key,) + keys
                yield path, value
        else:
            yield (key,), value



# -*- coding: utf-8 -*-
#
#    Monk is an unobtrusive data modeling, manipulation and validation library.
#    Copyright © 2011—2013  Andrey Mikhaylenko
#
#    This file is part of Monk.
#
#    Monk is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Monk is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with Monk.  If not, see <http://gnu.org/licenses/>.
"""
Monk
====

A simple schema validation layer for pymongo_. Inspired by MongoKit and Doqu.

.. _pymongo: http://api.mongodb.org/python/current/

"""
__version__ = '0.10.0'
# remember to also update:
#
# * PKGBUILD
#   * version
# * hg tag
#



# -*- coding: utf-8 -*-
#
#    Monk is an unobtrusive data modeling, manipulation and validation library.
#    Copyright © 2011—2013  Andrey Mikhaylenko
#
#    This file is part of Monk.
#
#    Monk is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Monk is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with Monk.  If not, see <http://gnu.org/licenses/>.
"""
Validation
~~~~~~~~~~
"""
from . import compat
from .schema import canonize
from . import errors

__all__ = [
    # functions
    'validate'
]



def validate_dict(rule, value):
    """ Nested dictionary. May contain complex structures which are validated
    recursively.

    The specification can be any dictionary, whether empty or not. It will be
    treated as a separate document.
    """
    validate_type(rule, value)

    if not rule.inner_spec:
        # spec is {} which means "a dict of anything"
        return

    spec_keys = set(rule.inner_spec.keys() if rule.inner_spec else [])
    data_keys = set(value.keys() if value else [])
    unknown = data_keys - spec_keys
    #missing = spec_keys - data_keys

    if unknown and not rule.dict_allow_unknown_keys:
        raise errors.UnknownKey('"{0}"'.format(
            '", "'.join(compat.safe_str(x) for x in unknown)))

    for key in spec_keys | data_keys:
        subrule = canonize(rule.inner_spec.get(key))
        if key in data_keys:
            value_ = value.get(key)
            try:
                validate(subrule, value_)
            except (errors.ValidationError, TypeError) as e:
                raise type(e)('{k}: {e}'.format(k=key, e=e))
        else:
            if subrule.optional:
                continue
            raise errors.MissingKey('{0}'.format(key))


def validate_list(rule, value):
    """ Nested list. May contain complex structures which are validated
    recursively.

    The specification can be either an empty list::

        >>> validate_list(Rule(list, inner_spec=[]), [123])

    ...or a list with exactly one item::

        >>> validate_list(Rule(list, inner_spec=[int]), [123, 456])
        >>> validate_list(Rule(list, inner_spec=[{'foo': int]), [{'foo': 123}])

    """
    if not isinstance(value, list):
        raise TypeError('expected {spec.__name__}, got '
                        '{valtype.__name__} {value!r}'.format(
                        spec=list, valtype=type(value),
                        value=value))

    if not rule.inner_spec:
        # spec is [] which means "a list of anything"
        return

    item_spec = canonize(rule.inner_spec)
    assert item_spec

    # XXX custom validation stuff can be inserted here, e.g. min/max items

    if not value and not item_spec.optional:
        raise errors.MissingValue('expected at least one item, got empty list')

    for i, item in enumerate(value):
        try:
            validate(item_spec, item)
        except (errors.ValidationError, TypeError) as e:
            raise type(e)('#{i}: {e}'.format(i=i, e=e))


def validate_type(rule, value):
    """ Simple type check.
    """
    if not isinstance(value, rule.datatype):
        raise TypeError('expected {typespec.__name__}, got '
                        '{valtype.__name__} {value!r}'.format(
                        typespec=rule.datatype, valtype=type(value),
                        value=value))



def validate(rule, value):
    """
    Validates given value against given specification.
    Raises an exception if the value is invalid.
    Always returns ``None``.

    :rule:
        a :class:`~monk.schema.Rule` instance or any other value digestible
        by :func:`canonize`.
    :value:
        any value including complex structures.

    Can raise:

    :class:`MissingValue`
        if a dictionary key is in the spec but not in the value.
        This applies to root and nested dictionaries.

    :class:`MissingKey`
        if a dictionary key is in the spec but not in the value.
        This applies to root and nested dictionaries.

    :class:`UnknownKey`
        if a dictionary key is the value but not not in the spec.

    :class:`StructureSpecificationError`
        if errors were found in spec.

    :class:`TypeError`
        if the value (or a nested value) does not belong to the designated type.

    """
    rule = canonize(rule)

    if value is None:
        # empty value, ok unless required
        if rule.optional:
            return

        if rule.datatype is None:
            raise errors.MissingValue('expected a value, got None')
        else:
            raise errors.MissingValue('expected {0}, got None'.format(rule.datatype.__name__))

    if rule.datatype is None:
        # any value is acceptable
        pass
    elif rule.datatype == dict:
        validate_dict(rule, value)
    elif rule.datatype == list:
        validate_list(rule, value)
    else:
        assert not rule.inner_spec
        if isinstance(rule.datatype, type):
            validate_type(rule, value)

    for validator in rule.validators:
        validator(value)
